from gitsub.display import show
from gitsub.repo import RepoFactory
from gitsub.env import Environment

def handler(args, env):
    repolist = env.cache["repolist"]
    repos = [RepoFactory.new(path) for path in repolist]
    updates = [repo.query(["name", "commits"]) for repo in repos ]
    show(updates)




import os

class Cache(dict):
    DEFAULT_CACHE_DIR = os.environ.get(
        "GITSUB_CACHE_DIR",
        os.path.expanduser("~/.cache/gitsub/")
    )

    DEFAULT_CACHES = [ "repolist" ]

    def __init__(self, directory=DEFAULT_CACHE_DIR, caches=DEFAULT_CACHES):
        self.directory = directory
        self.caches = caches

    def exists(self):
        return (os.path.exists(self.directory) and
                all(os.path.exists(os.path.join(self.directory, cache))
                    for cache in self.caches)
                )

    def write(self):
        if not os.path.exists(self.directory):
            os.makedirs(self.directory, exist_ok=True)

        for cache in self.caches:
            cache_file = os.path.join(
                self.directory,
                cache
            )

            with open(cache_file, "w") as f:
                if cache in self:
                    print("\n".join(self[cache]), file=f)

    def read(self):
        if not os.path.exists(self.directory):
            raise IOError()

        for cache in self.caches:
            cache_file = os.path.join(
                self.directory,
                cache
            )

            try:
                self[cache] = open(cache_file,"r").read().splitlines()
            except IOError as ex:
                raise



import os
from gitsub.git import is_git

def handler(args, env):
    paths = args
    repolist = env.cache["repolist"]
    for path in paths:
        new_repos = traverse(os.path.abspath(path))
        for new_path in new_repos:
            if not new_path in repolist:
                repolist.append(new_path)
    env.cache.write()

def traverse(path):
    if is_git(path):
        return [path]
    else:
        new_repos = []
        dirents = os.scandir(path)
        for entry in dirents:
            if entry.is_dir():
                new_repos += traverse(entry.path)
        return new_repos



import argparse
import sys

import gitsub.add as add
import gitsub.ls as ls
import gitsub.fetch as fetch
import gitsub.sync as sync
from gitsub.utils import get_version
from gitsub.env import Environment

handler_list = {
    "add": add.handler,
    "ls": ls.handler,
    "fetch": fetch.handler,
    "sync": sync.handler,
}

def run_args(args):
    command = args.command
    command_args = args.command_args
    cmd_handler = handler_list[command]
    env = Environment(cache_dir=args.cache_dir)
    cmd_handler(command_args,env)


def main():
    parser = argparse.ArgumentParser(prog="gitsub")
    parser.add_argument("-v", "--version",
                        action="version",
                        version="gitsub v{}".
                        format(get_version())
                       )
    parser.add_argument("--cache-dir",
                        metavar="PATH",
                        default=Environment.cache_dir,
                        help="set custom repositories catalog"
                       )
    parser.add_argument("command",
                        choices=handler_list.keys(),
                        nargs="?",
                        default="fetch"
                       )
    parser.add_argument("command_args",
                        nargs=argparse.REMAINDER,
                        )

    args = parser.parse_args(sys.argv[1:])
    run_args(args)



def show(repos):
    for repo in repos:
        print("'{}' repository updates".format(repo['name']))
        updates = repo['commits']
        if updates != None:
            for update in updates:
                print("\t"+update)




from gitsub.repo import RepoFactory
from gitsub.env import Environment

def handler(args, env):
    repolist = env.cache["repolist"]
    repos = [RepoFactory.new(path) for path in repolist]
    for repo in repos:
        repo.sync()
    print("repositories synced")





import os
import gitsub.git as git

class RepoFactory:
    @staticmethod
    def new(path):
        if (git.is_git(path)):
            return Repo(path)
        else:
            return None

class Repo:
    def __init__(self, path):
        self.path = os.path.normpath(path)
        self.name = os.path.basename(self.path)
        self.branches = git.get_branches(self.path)
        self.cur_branch = git.get_branch(self.path)


    def query(self, queries):
        return {query:self.handle_query(query) for query in queries}

    def handle_query(self, query):
        if query == "name":
            return self.name
        if query == "commits":
            return self.get_updates()

    def sync(self):
        p = git.update_remotes(self.path)
        return p != None

    def get_updates(self):
        for branch in self.branches:
            local = git.branch_local_head(self.path, branch)
            remote = git.branch_remote_head(self.path, branch)
            base = git.branch_merge_base(self.path, branch)
            if base == None:
                print(self.path)
            else:
                if local == remote:
                    return None
                elif local == base:
                    # remote is up-to-date
                    commits = git.get_commits(self.path, ["^" + base,remote])
                    return commits
                elif remote == base:
                    # local is up-to-date
                    commits = git.get_commits(self.path, ["^" + base,local])
                    return commits
                else:
                    new_commits = git.get_commits(self.path, ["^" + base,local])
                    new_updates = git.get_commits(self.path, ["^" + base,remote])
                    return new_commits + new_updates



import os
from gitsub.cache import Cache

class Environment(object):
    cache_dir = Cache.DEFAULT_CACHE_DIR

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    @property
    def cache(self):
        if not hasattr(self, "_cache"):
            self._cache = Cache(directory=self.cache_dir)
            if not self._cache.exists():
                self._cache.write()
            self._cache.read()
        return self._cache

    def cleanup(self):
        self._cache.write()

    def __del__(self):
        try:
            self.cleanup()
        except Exception:
            pass



from setuptools import setup, find_packages
import pypandoc

import gitsub

long_description = pypandoc.convert('README.md','rst')

setup(
    name="gitsub",
    version=gitsub.__version__,
    description="Track git repositories updates",
    long_description=long_description,
    author="Aufar Gilbran",
    author_email="aufargilbran@gmail.com",
    url="https://github.com/aufarg/gitsub",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Version Control"
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "gitsub = gitsub.main:main",
        ]
    }

)



import subprocess
from gitsub.utils import sub_output

def _run(path, cmd, opt=[], args=[],
         check=True,
         universal_newlines=True,
         stdout=subprocess.PIPE,
         stderr=subprocess.DEVNULL,
        ):
    git_cmd = cmd + opt + args
    sh_cmd = ["git", "-C", path] + git_cmd
    return subprocess.run(sh_cmd,
                          check=check,
                          stdout=stdout,
                          stderr=stderr,
                          universal_newlines=universal_newlines
                         )

def branch_local_head(path, branch):
    return _rev_parse(path,branch+"@{0}")

def branch_remote_head(path, branch):
    return _rev_parse(path,branch+"@{u}")

def branch_merge_base(path, branch):
    cmd = ["merge-base", branch+"@{0}", branch+"@{u}"]
    try:
        return sub_output(_run(path, cmd))
    except subprocess.CalledProcessError as ex:
        return None

def _rev_parse(path, interval):
    cmd = ["rev-parse", interval]
    try:
        return sub_output(_run(path, cmd))
    except subprocess.CalledProcessError as ex:
        return []

def get_commits(path, args):
    cmd = ["rev-list"]
    opt = ["--max-count=3", "--oneline"]
    try:
        return sub_output(_run(path, cmd, opt, args)).splitlines()
    except subprocess.CalledProcessError as ex:
        return []

def get_branch(path):
    cmd = ["symbolic-ref", "HEAD"]
    refs_prefix = "refs/heads/"
    try:
        ref = sub_output(_run(path, cmd))
        if ref.startswith(refs_prefix):
            return ref[len(refs_prefix):]
        else:
            return None
    except subprocess.CalledProcessError as ex:
        return None

def get_branches(path):
    try:
        cmd = ["for-each-ref"]
        opt = ["--format=%(refname:short)"]
        args = ["refs/heads/"]
        return sub_output(_run(path, cmd, opt, args)).split("\n")
    except subprocess.CalledProcessError as ex:
        return []

def is_git(path):
    try:
        p = _run(path, ["status"])
        return True
    except subprocess.CalledProcessError:
        return False

def update_remotes(path):
    try:
        p = _run(path, ["remote", "update"], stdout=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False
